"""
Simple server for loading d3 json files
Adapted from networkx example at:
    https://github.com/networkx/networkx/blob/master/examples/javascript/http_server.py
"""
import SimpleHTTPServer, BaseHTTPServer
import socket
import thread
import webbrowser
import os

def serve(directory, filename, port):
    base_url = "127.0.0.1"
    os.chdir(directory)
    httpd = StoppableHTTPServer(
        (base_url, port), 
        SimpleHTTPServer.SimpleHTTPRequestHandler
    )
    thread.start_new_thread(httpd.serve, ())
    webbrowser.open_new('http://%s:%s/%s' % (base_url, port, filename))
    raw_input("Press <RETURN> to stop server\n")
    httpd.stop()


class StoppableHTTPServer(BaseHTTPServer.HTTPServer):

    def server_bind(self):
        BaseHTTPServer.HTTPServer.server_bind(self)
        self.socket.settimeout(1)
        self.run = True

    def get_request(self):
        while self.run:
            try:
                sock, addr = self.socket.accept()
                sock.settimeout(None)
                return (sock, addr)
            except socket.timeout:
                pass

    def stop(self):
        self.run = False

    def serve(self):
        while self.run:
            self.handle_request()



import os
import ast
import itertools

from blackwidow.error import BlackWidowIOError, BlackWidowValueError


class Node(object):
    """
    A node in the python project graph.
    Contains a list of outgoing edges.
    """

    def __init__(self, filename, package_root=None):
        if not package_root:
            raise SpiderValueError("No 'package_root' specified.")
        self.filename = filename
        self.package_root = package_root
        self.module = import_path(filename, os.path.dirname(package_root))
        self.edges = imported_modules(filename, package_root=package_root)
        self.imports = list(itertools.chain(*[edge.imports for edge in self.edges]))


class Import(object):
    """
    Standardizes ast.Import and ast.ImportFrom objects into a common format
    """

    def __init__(self, obj, filename, package_root):
        self.raw = obj
        self.imports = []

        if hasattr(obj, 'module') and obj.module != None:
            self.module = obj.module
        elif hasattr(obj, 'level'):
            self.module = import_path(filename, package_root, level=obj.level)
        else:
            self.module = None

        for alias in obj.names:
            self.imports.append(self._absolute_import(alias))

    def _absolute_import(self, alias):
        if not self.module:
            return alias.name
        return '.'.join([self.module, alias.name])


def import_path(filename, package_root, level=None):
    """
    Given a filename within a package, return the import path of the filename
    """
    relative_path = os.path.relpath(filename, package_root)
    path = os.path.splitext(relative_path)[0]
    dirs = path.split(os.path.sep)
    if level != None:
        dirs = dirs[:-level]
    return ".".join(dirs)


def imported_modules(filename, package_root):
    """
    Given a filename, returns a list of modules imported by the file
    """
    try:
        fd = open(filename, 'rt')
        tree = ast.parse(fd.read(), filename=filename)
        fd.close()
    except IOError as e:
        raise SpiderIOError(e)

    imports = []
    for item in tree.body:
        if isinstance(item, (ast.Import, ast.ImportFrom)):
            imports.append(
                Import(item, filename=filename, package_root=package_root)
            )
    return imports



from setuptools import setup, find_packages

setup(
    name="blackwidow",
    version="0.1.0",
    author="Madison May",
    author_email="madison@indico.io",
    packages=find_packages(),
    install_requires=[
        "networkx==1.10"
    ],
    description="""
        Visualizing and refactoring python project import graphs.
    """,
    license="MIT License (See LICENSE)",
    long_description=open("README.rst").read(),
    url="https://github.com/madisonmay/spider"
)


import os
import json

import networkx as nx
from networkx.readwrite import json_graph

import blackwidow.config as config
from blackwidow.node import Node
from blackwidow.viz.server import serve


class Web(object):
    """
    The full project web.  
    Contains the package root and list of nodes.
    """

    def __init__(self, package_root, exclude=None):
        files = find_files(package_root, exclude=exclude)
        self.package_root = package_root
        self.nodes = [
            Node(file, package_root=self.package_root) for file in files
        ]

        self.graph = nx.Graph()

        for node in self.nodes:
            self.graph.add_node(node.module)
            self.graph.node[node.module]['name'] = node.module

        for node in self.nodes:
            for module in node.imports:
                self.graph.add_edge(node.module, module)

    def visualize(self, port=None):
        """
        Dumps graph data to a json file and renders using d3's force-directed graph
        """
        json_data = json_graph.node_link_data(self.graph)

        parent_dir = os.path.abspath(os.path.dirname(__file__))
        viz_dir = os.path.join(parent_dir, 'viz', 'static')
        json_file = os.path.join(viz_dir, 'graph.json')
        html_file = os.path.join(viz_dir, 'graph.html')
        relative_path = os.path.relpath(html_file, viz_dir)

        with open(json_file, 'w') as fd:
            json.dump(json_data, fd)

        if not port:
            port = config.PORT

        serve(directory=viz_dir, filename=relative_path, port=port)


def find_files(dir, exclude=None):
    """
    Given a root directory, compiles a list of python files to analyze.
    If a discovered python file is provided in the list of excluded, it will be ignored
    """
    if not exclude:
        exclude = []

    matches = []
    for root, dirs, files in os.walk(dir):
        for file in files:
            if file.endswith(".py") and not file in exclude:
                matches.append(
                    os.path.abspath(os.path.join(root, file))
                )

    return matches


if __name__ == "__main__":
    import sys
    path = sys.argv[1]
    web = Web(path)
    web.visualize()
